#define _XOPEN_SOURCE 1
#define _XOPEN_SOURCE_EXTENDED 1
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <limits.h>
#include <time.h>
#include <signal.h>

#define PROG "somewhen"

static int seen_any=0;
static int increased=0;
static int need_increase=0;

static char const var_lastrun[]="$LASTRUN";
static char const var_maxinterval[]="$MAXINTERVAL";
static char const var_weekday[]="$WEEKDAY";
static char const var_year[]="$YEAR";
static char const var_month[]="$MONTH";
static char const var_monthday[]="$MONTHDAY";
static char const var_hour[]="$HOUR";
static char const var_minute[]="$MINUTE";
static char const var_second[]="$SECOND";
static char const var_printdelay[]="$PRINTDELAY";

struct caltime ct;

static unsigned long limit_weekday(void) { return  7; }
static unsigned long limit_year   (void) { return -1; }
static unsigned long limit_month  (void) { return 12; }
static unsigned long limit_monthday(void) {
  static unsigned long const limits[]={
    31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31
  };
  unsigned long leap;
  if (ct.date.month!=1) return limits[ct.date.month];
  leap=ct.date.year;
  if (leap%4!=0) leap=0;
  else {
    leap/=4;
    if (leap%25!=0) leap=1;
    else {
      leap/=25;
      if (leap%4!=0) leap=0;
      else leap=1;
    }
  }
  return limits[ct.date.month]+leap;
}
static unsigned long limit_hour  (void) { return 24; }
static unsigned long limit_minute(void) { return 60; }
static unsigned long limit_second(void) { return 60; }

typedef enum { unset, eq, div } spectype;
typedef struct {
  char const* const env;
  spectype type;
  unsigned long spec;
  unsigned long value;
  unsigned long (* const limit)(void);
} unit;
static unit weekday_units[]={
  { var_weekday,  -1, -1, -1, limit_weekday  },
  { var_hour,     -1, -1, -1, limit_hour     },
  { var_minute,   -1, -1, -1, limit_minute   },
  { var_second,   -1, -1, -1, limit_second   }
};
static unit monthday_units[]={
  { var_year,     -1, -1, -1, limit_year     },
  { var_month,    -1, -1, -1, limit_month    },
  { var_monthday, -1, -1, -1, limit_monthday },
  { var_hour,     -1, -1, -1, limit_hour     },
  { var_minute,   -1, -1, -1, limit_minute   },
  { var_second,   -1, -1, -1, limit_second   }
};

static void do_unit(unit* u) {
  unsigned long value;
  unsigned long spec;
  spectype type;
  if (increased) value=0;
  else value=u->value;
  type=u->type;
  spec=u->spec;
  if (type!=unset) seen_any=1;
  else {
    if (need_increase) {
      ++value;
      increased=1;
      need_increase=0;
      goto setval;
    }
    if (!seen_any) return;
    if (increased) goto setval; /* shortcut */
    type=eq;
    spec=0;
  }
  if (type==eq) {
    if (need_increase) return;
    if (value==spec) goto setval;
    if (value>spec) goto need_increase;
    value=spec;
  } else /* type==div */ {
    if (need_increase) {
      value+=spec;
      if (value>=u->limit()) return;
      increased=1;
      need_increase=0;
      goto setval;
    }
    if (value==0) goto setval;
    if (spec==0) goto need_increase;
    if (value%spec==0) goto setval;
    value+=spec;
    value-=value%spec;
    if (value>=u->limit()) goto need_increase;
  }
  increased=1;
  setval:
  u->value=value;
  return;
  need_increase:
  need_increase=1;
  return;
}

static void alarm_handler(int sig) {
  (void)sig;
}

int main(int argc, char** argv) {
  struct tai now;
  unsigned long timenow;
  unsigned long delay;
  int weekday;
  unsigned int i;
  char const* x;
  unit* units;
  unsigned int n_units;

  x=envget(var_weekday);
  if (x) {
    units=weekday_units;
    n_units=sizeof weekday_units/sizeof (unit);
  } else {
    units=monthday_units;
    n_units=sizeof monthday_units/sizeof (unit);
  }

  for (i=0; i!=n_units; ++i) {
    x=envget(units[i].env);
    if (x==null) units[i].type=unset;
    else {
      unsigned int len;
      seen_any=1;
      switch (*x) {
        case '=': units[i].type=eq;  break;
        case '/': units[i].type=div; break;
        default:
          die3(x, 111, FATAL, units[i].env, err_malformed);
      }
      len=scan_ulong(x+1, &units[i].spec);
      if (len==0 || x[1+len]!='\0')
        die3(x, 111, FATAL, units[i].env, err_malformed);
    }
  }

  if (!seen_any)
    die2(x, 111, FATAL, "no run times specified");
  seen_any=0;

  tai_now(&now);
  timenow=time(null);
  caltime_utc(&ct, &now, &weekday, null);

  if (units==weekday_units) {
    units[0].value=weekday;
    units[1].value=ct.hour;
    units[2].value=ct.minute;
    units[3].value=ct.second;
  } else {
    units[0].value=ct.date.year;
    units[1].value=ct.date.month-1;
    units[2].value=ct.date.day-1;
    units[3].value=ct.hour;
    units[4].value=ct.minute;
    units[5].value=ct.second;
  }

  for (i=0; i!=n_units;) {
    do_unit(&units[i]);
    if (need_increase) {
      for (;;) {
        if (i==0) {
          if (units==weekday_units) {
            units[0].value=0;
            ct.date.day+=7;
            continue;
          }
          die2(x, 111, FATAL, "all run times are in the past");
        }
        --i;
        do_unit(&units[i]);
        if (increased) break;
      }
    }
    ++i;
  }

  if (units==weekday_units) {
    ct.date.day +=units[0].value;
    ct.date.day -=weekday;
    ct.hour      =units[1].value;
    ct.minute    =units[2].value;
    ct.second    =units[3].value;
  } else {
    ct.date.year =units[0].value;
    ct.date.month=units[1].value+1;
    ct.date.day  =units[2].value+1;
    ct.hour      =units[3].value;
    ct.minute    =units[4].value;
    ct.second    =units[5].value;
  }

  {
    unsigned char buf[TAI_PACK];
    static char const overflow[]="delay overflows long";
    struct tai then;
    caltime_tai(&ct, &then);
    tai_sub(&then, &then, &now);
    tai_pack(buf, &then);
    delay=0;
    for (i=0; i<TAI_PACK; ++i) {
      unsigned long const d=delay<<CHAR_BIT;
      if (d>>CHAR_BIT!=delay) die2(x, 111, FATAL, overflow);
      delay=d+buf[i];
      if (delay<d) die2(x, 111, FATAL, overflow);
    }
    if (delay>LONG_MAX) die2(x, 111, FATAL, overflow);
  }

  x=envget(var_maxinterval);
  if (x!=null) {
    struct stat statbuf;
    unsigned long max;
    i=scan_ulong(x, &max);
    if (i==0 || x[1+i]!='\0')
      die3(x, 111, FATAL, var_maxinterval, err_malformed);
    x=envget(var_lastrun);
    if (x==null) x=PROG "_lastrun";
    if (stat(x, &statbuf)==0) {
      unsigned long mtime=statbuf.st_mtime;
      if (timenow+delay>mtime+max)
        delay=mtime+max-timenow;
    } else if (errno!=ENOENT)
      die4(sys, 111, FATAL, err_stat, x, err);
  }

  if (delay==0) delay=1;

  if (envget(var_printdelay)!=null) {
    char buf[(sizeof delay*CHAR_BIT+2)/3+1];
    buf[fmt_ulong(buf, delay)]='\0';
    strerr_warn1(buf, null);
  }

  {
    struct itimerval timer;
    sigset_t oldset, alarmset, suspendset;
    struct sigaction oldact, alarmact;
    timer.it_interval.tv_sec=0;
    timer.it_interval.tv_usec=0;
    timer.it_value.tv_sec=delay;
    timer.it_value.tv_usec=0;
    if (sigemptyset(&alarmset)!=0)
      die2(sys, 111, FATAL, err_sigempty);
    if (sigaddset(&alarmset, SIGALRM)!=0)
      die2(sys, 111, FATAL, err_sigadd);
    sigprocmask(SIG_BLOCK, null, &suspendset);
    sigprocmask(SIG_BLOCK, &alarmset, &oldset);
    setitimer(ITIMER_REAL, &timer, null);
    alarmact.sa_handler=alarm_handler;
    if (sigemptyset(&alarmact.sa_mask)!=0)
      die2(sys, 111, FATAL, err_sigempty);
    alarmact.sa_flags=0;
    if (sigaction(SIGALRM, &alarmact, &oldact)!=0)
      die2(sys, 111, FATAL, err_sigact);
    if (sigdelset(&suspendset, SIGALRM)!=0)
      die2(sys, 111, FATAL, err_sigdel);
    sigsuspend(&suspendset);
    if (sigaction(SIGALRM, &oldact, null)!=0)
      die2(sys, 111, FATAL, err_sigact);
    sigprocmask(SIG_SETMASK, &oldset, null);
  }

  if (x!=null) {
    int fd=open(x, O_WRONLY|O_CREAT, 0600);
    if (fd<0)
      die4(sys, 111, FATAL, err_create, x, err);
    if (close(fd)!=0)
      die4(sys, 111, FATAL, err_close, x, err);
  }

  if (argc<=1) _exit(0);

  ++argv;
  pathexec(argv);
  die4(sys, 111, FATAL, err_exec, *argv, err);
}
